/***************************************************************************
 *   Copyright (C) 2008 by Lisandro Damián Nicanor Pérez Meyer             *
 *   perezmeyer@gmail.com                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "dataholder.h"

#include <QList>
#include <QDebug>

DataHolder::DataHolder(int rows, int columns)
{
	numberOfRows = rows;
	numberOfColumns = columns;
	matrix = new Point4F[rows * columns];
}

DataHolder::~DataHolder()
{
	delete [] matrix;
}

void DataHolder::addData(int row, int column, const Point4F & data)
{
	if(row < 0 || row >= numberOfRows)
	{
		qDebug() << __func__ <<  QObject::tr("row is out of bounds. It's value is ") << row;
		return;
	}
	if(column < 0 || row >= numberOfColumns)
	{
		qDebug() << __func__ << QObject::tr("column is out of bounds. It's value is ") << column;
		return;
	}
	matrix[row * numberOfColumns + column] = data;
}

QVector<double> * DataHolder::getVertexArray(int row) const
{
	if(row < 0 || row >= numberOfRows -1)
	{
		qDebug() << __func__ << QObject::tr("row is out of bounds. It's value is ") << row;
		return 0;
	}

	QVector<double> * va = new QVector<double>;

	for(int column=0; column < numberOfColumns; column++)
	{
		va->append(matrix[row * numberOfColumns + column].x());
		va->append(matrix[row * numberOfColumns + column].y());
		va->append(matrix[row * numberOfColumns + column].z());

		va->append(matrix[(row + 1) * numberOfColumns + column].x());
		va->append(matrix[(row + 1) * numberOfColumns + column].y());
		va->append(matrix[(row + 1) * numberOfColumns + column].z());
	}

	return va;
}

QVector<double> * DataHolder::getNormalArray(int row) const
{
	Point4F first(0.0,0.0,0.0);
	Point4F second(0.0,0.0,0.0);
	Point4F temp(0.0,0.0,0.0);

	if(row < 0 || row >= numberOfRows -1)
	{
		qDebug() << __func__ << QObject::tr("row is out of bounds. It's value is ") << row;
		return 0;
	}

	QVector<double> * na = new QVector<double>;

	/*
	On calculating the normals we have five special cases:
		* The first row has to access the last one as it's upper row
		* The last vertex of each row has to access the first vertex
		* The last vertex of the first row has the two problems described before
		* The las row has to access the first row
		* The last vertex of the last row has the problem above and also has to
		  access the first vertex

	The cross products will be calculated using as the first vector the one formed
	between the current vertex and the following, and as the second vector the one
	formed with the current vertex and the upper one.
	*/

	// The first row
	if(row == 0)
	{
		// First we do all the vertex except the last one
		for(int column=0; column < numberOfColumns - 1; column++ )
		{
			// The vertex
			first = matrix[column + 1] -
			        matrix[column];
			second = matrix[numberOfColumns + column] -
			         matrix[column];
			temp = calculateNormal(first, second);

			na->append(temp.x());
			na->append(temp.y());
			na->append(temp.z());

			// The vertex below
			first = matrix[numberOfColumns + column + 1] -
			        matrix[numberOfColumns + column];
			second = matrix[column] -
			         matrix[numberOfColumns + column];
			temp = calculateNormal(second, first);

			na->append(temp.x());
			na->append(temp.y());
			na->append(temp.z());
		}
		// The last vertex
		first = matrix[0] -
		        matrix[numberOfColumns - 1];
		second = matrix[numberOfColumns + numberOfColumns - 1] -
		         matrix[numberOfColumns - 1];
		temp = calculateNormal(second, first);

		na->append(temp.x());
		na->append(temp.y());
		na->append(temp.z());

		// The last vertex below
		first = matrix[numberOfColumns] -
		        matrix[numberOfColumns + numberOfColumns -1];
		second = matrix[numberOfColumns -1] -
		         matrix[numberOfColumns + numberOfColumns -1];
		temp = calculateNormal(first, second);

		na->append(temp.x());
		na->append(temp.y());
		na->append(temp.z());
	}
	// The common cases
	else if(row > 0 && row < numberOfRows -1)
	{
		// First we do all the vertex except the last one
		for(int column=0; column < numberOfColumns - 1; column++ )
		{
			// The vertex
			first = matrix[row * numberOfColumns + column + 1] -
			        matrix[row * numberOfColumns + column];
			second = matrix[(row - 1) * numberOfColumns + column] -
			         matrix[row * numberOfColumns + column];
			temp = calculateNormal(second, first);

			na->append(temp.x());
			na->append(temp.y());
			na->append(temp.z());

			// The vertex below
			first = matrix[(row + 1) * numberOfColumns + column + 1] -
			        matrix[(row + 1) * numberOfColumns + column];
			second = matrix[row * numberOfColumns + column] -
			         matrix[(row + 1) * numberOfColumns + column];
			temp = calculateNormal(second, first);

			na->append(temp.x());
			na->append(temp.y());
			na->append(temp.z());
		}
		// The last vertex
		first = matrix[row * numberOfColumns] -
		        matrix[row * numberOfColumns + numberOfColumns - 1];
		second = matrix[(row - 1) * numberOfColumns + numberOfColumns -1] -
		         matrix[row * numberOfColumns + numberOfColumns - 1];
		temp = calculateNormal(first, second);

		na->append(temp.x());
		na->append(temp.y());
		na->append(temp.z());

		// The last vertex below
		first = matrix[(row + 1) * numberOfColumns] -
		        matrix[(row + 1) * numberOfColumns + numberOfColumns - 1];
		second = matrix[row * numberOfColumns + numberOfColumns -1] -
		         matrix[(row + 1) * numberOfColumns + numberOfColumns - 1];
		temp = calculateNormal(first, second);

		na->append(temp.x());
		na->append(temp.y());
		na->append(temp.z());
	}
	// The last row
	else
	{
		// First we do all the vertex except the last one
		for(int column=0; column < numberOfColumns - 1; column++ )
		{
			// The vertex
			first = matrix[row * numberOfColumns + column + 1] -
			        matrix[row * numberOfColumns + column];
			second = matrix[(row - 1) * numberOfColumns + column] -
			         matrix[row * numberOfColumns + column];
			temp = calculateNormal(first, second);

			na->append(temp.x());
			na->append(temp.y());
			na->append(temp.z());

			// The vertex below (which happens to be the first row)
			first = matrix[column + 1] -
			        matrix[column];
			second = matrix[row * numberOfColumns + column] -
			         matrix[column];
			temp = calculateNormal(first, second);

			na->append(temp.x());
			na->append(temp.y());
			na->append(temp.z());
		}
		// The last vertex
		first = matrix[row * numberOfColumns] -
		        matrix[row * numberOfColumns + numberOfColumns - 1];
		second = matrix[(row - 1) * numberOfColumns + numberOfColumns -1] -
		         matrix[row * numberOfColumns + numberOfColumns - 1];
		temp = calculateNormal(first, second);

		na->append(temp.x());
		na->append(temp.y());
		na->append(temp.z());

		// The last vertex below (which happens to be in the first row)
		first = matrix[0] -
		        matrix[numberOfColumns - 1];
		second = matrix[row * numberOfColumns + numberOfColumns - 1] -
		         matrix[numberOfColumns - 1];
		temp = calculateNormal(first, second);

		na->append(temp.x());
		na->append(temp.y());
		na->append(temp.z());
	}

	return na;
}

Point4F DataHolder::calculateNormal(Point4F first, Point4F second) const
{
	double x(0.0), y(0.0), z(0.0);

	x = first.y()*second.z() - first.z()*second.y();
	y = first.z()*second.x() - first.x()*second.z();
	z = first.x()*second.y() - first.y()*second.x();

	return Point4F(x, y, z);
}

