/***************************************************************************
 *   Copyright (C) 2005 by                                                 *
 *   Lisandro Damián Nicanor Pérez Meyer - perezmeyer en/at gmail.com      *
 *   Gustavo González - gonzalgustavo en/at gmail.com                      *
 *   Pablo Odorico  pablo.odorico en/at gmail.com                          *
 *                                                                         *
 *   Modified by Lisandro Damián Nicanor Pérez Meyer on February 2008 for  *
 *   a project for the subject "Programación de sistemas gráficos"         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// GLEW must be included first or it complains
#include <GL/glew.h>
#include <QString>
#include <QMessageBox>
#include <QMouseEvent>
#include <QPixmap>
#include <QFile>
#include <QByteArray>
#include <QDebug>
#include <math.h>
#include "glslutils.h"
#include "glwidget.h"

#define POW2(a)         ((a)*(a))
#define POW3(a)         ((a)*(a)*(a))
#define PI 3.141592654

GLWidget::GLWidget(QWidget * parent)
	: QGLWidget(parent)
{
	drawAxis = false;
	drawNormals = false;
	drawWireframe = false;
	useHalfway = false;

	currentMethod = Constants::Bezier;
	currentFigure = Constants::Patch;

	lightPosition[0] = 1.0;
	lightPosition[1] = 0.5;
	lightPosition[2] = 1.0;
	lightPosition[3] = 0.0;

	xRot = 0;
	yRot = 0;
	zRot = 0;

	currentTexture = 0;
}

GLWidget::~GLWidget()
{
	vertexArray->clear();
	delete vertexArray;

	qDeleteAll(holdersList.begin(), holdersList.end());
	holdersList.clear();
}

void GLWidget::setDrawAxis(bool draw)
{
	drawAxis = draw;
	updateGL();
}

void GLWidget::setDrawNormals(bool draw)
{
	drawNormals = draw;
	updateGL();
}

void GLWidget::setDrawWireframe(bool draw)
{
	drawWireframe = draw;
	updateGL();
}

void GLWidget::setFigure(Constants::Figures figure)
{
	currentFigure = figure;
	calculateSurface();
	update();
}

void GLWidget::setMethod(Constants::Methods method)
{
	currentMethod = method;
	calculateSurface();
	update();
}

void GLWidget::setXRotation(int angle)
{
	normalizeAngle(&angle);
	if(angle != xRot)
		xRot = angle;
}

void GLWidget::setYRotation(int angle)
{
	normalizeAngle(&angle);
	if(angle != yRot)
		yRot = angle;
}

void GLWidget::setZRotation(int angle)
{
	normalizeAngle(&angle);
	if(angle != zRot)
		zRot = angle;
}

void GLWidget::setUseShaders(int use)
{
	if(use == Qt::Checked)
		setShaders();
	else
		unsetShaders();
	updateGL();
}

void GLWidget::setUseHalfway(int use)
{
	if(use == Qt::Checked)
		useHalfway = true;
	else
		useHalfway = false;

	unsetShaders();
	setShaders();

	updateGL();
}

void GLWidget::setTexture(int texture)
{
	if(texture >= 0 && texture < maxTextures)
		currentTexture = texture;
	updateGL();
}

void GLWidget::setXLightPosition(int position)
{
	lightPosition[0] = (float)(position/10.0);
	updateGL();
}

void GLWidget::setYLightPosition(int position)
{
	lightPosition[1] = (float)(position/10.0);
	updateGL();
}

void GLWidget::setZLightPosition(int position)
{
	lightPosition[2] = (float)(position/10.0);
}

void GLWidget::paintGL()
{
	// Texture binding
	texName = bindTexture(QPixmap(QString(":/images/image%1.png").arg(currentTexture)),
	                      GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texName);


	glClearColor(135.0/255, 135.0/255, 135.0/255, 1);

	// Clean
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();

	// Put your lights on ;-)
	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);


	// Rotate the camera
	glTranslatef(0.0, 0.0, -10.0);
	glRotated(xRot / 16.0, 1.0, 0.0, 0.0);
	glRotated(yRot / 16.0, 0.0, 1.0, 0.0);
	glRotated(zRot / 16.0, 0.0, 0.0, 1.0);

	// Coordinates for the camera
	double x(0.0), y(0.0), z(0.0);

	// ID for the camera position uniform variable
	GLint cameraPositionID;
	cameraPositionID = glGetUniformLocationARB(p, "cameraPosition");

	// Calculate the camera position
	x = sin(-yRot * (PI/180.0));
	z = cos(-yRot * (PI/180.0));
	y = tan(xRot * (PI/180.0));

	// Attach the camera position to the shader
	glUniform3fARB(cameraPositionID, x, y, z);

	// Axis
	if(drawAxis)
	{
		glLineWidth(1);
		glBegin(GL_LINES);
			// Red, X axis
			glColor3f(1.0, 0.0, 0.0);
			glVertex3f(0.0, 0.0, 0.0);
			glVertex3f(1.0, 0.0, 0.0);

			// Green, Y axis
			glColor3f(0.0, 1.0, 0.0);
			glVertex3f(0.0, 0.0, 0.0);
			glVertex3f(0.0, 1.0, 0.0);

			// Blue, Z axis
			glColor3f(0.0, 0.0, 1.0);
			glVertex3f(0.0, 0.0, 0.0);
			glVertex3f(0.0, 0.0, 1.0);
		glEnd();
		glLineWidth(0);
	}

	for(int currentHolder=0; currentHolder < holdersList.size(); currentHolder++)
	{
		for(int i=0; i < holdersList.at(currentHolder)->rows() - 1; i++)
		{
			// We enable the arrays
			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_NORMAL_ARRAY);

			vertexArray = holdersList.at(currentHolder)->getVertexArray(i);
			normalArray = holdersList.at(currentHolder)->getNormalArray(i);
			vectorNormalArray = getNormalVectors(vertexArray, normalArray);

			if(vertexArray->size() > 3)
			{
				glVertexPointer(3,GL_DOUBLE,0,vertexArray->data());
				glNormalPointer(GL_DOUBLE,0,normalArray->data());

				// We draw the arrays.
				// We define a different color every three holders
				if((currentHolder % 3) == 0)
					glColor4f(1.0, 0.0, 0.0, 1.0);
				else if((currentHolder % 3) == 1)
					glColor4f(0.0, 1.0, 0.0, 1.0);
				else
					glColor4f(0.0, 0.0, 1.0, 1.0);

				glDrawArrays(GL_TRIANGLE_STRIP,0, vertexArray->size()/3);

				if(drawWireframe)
				{
					// To draw the wireframe, we use a white line
					glColor3f(1.0, 1.0, 1.0);
					glDrawArrays(GL_LINE_STRIP,0, vertexArray->size()/3);
				}

				// Disable the normals
				glDisableClientState(GL_NORMAL_ARRAY);


				if(drawNormals)
				{
					// Draw the normal vectors
					// We define a different color every three holders
					if((currentHolder % 3) == 0)
						glColor4f(1.0, 0.0, 0.0, 1.0);
					else if((currentHolder % 3) == 1)
						glColor4f(0.0, 1.0, 0.0, 1.0);
					else
						glColor4f(0.0, 0.0, 1.0, 1.0);

					glVertexPointer(3,GL_DOUBLE,0,vectorNormalArray->data());
					glDrawArrays(GL_LINES,0, vectorNormalArray->size()/3);
				}
			}
			delete vertexArray;
			delete normalArray;
			delete vectorNormalArray;
		}
	}

	swapBuffers();
}

void GLWidget::resizeGL(int width, int height)
{
	int side = qMin(width, height);
	glViewport((width - side) / 2, (height - side) / 2, side, side);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, 1.0, -1.0, -4.0, 150.0);
	glMatrixMode(GL_MODELVIEW);
}

void GLWidget::initializeGL()
{
	emit lightXPositionChanged((int)(lightPosition[0]*10));
	emit lightYPositionChanged((int)(lightPosition[1]*10));
	emit lightZPositionChanged((int)(lightPosition[2]*10));

	// Check wheter we have OpenGL support or not
	if(!QGLFormat::hasOpenGL())
	{
		qDebug() << __func__ << QObject::tr("We don't have OpenGL support");
		QMessageBox::critical(0,QObject::tr("This beautiful program"),
		                      QObject::tr("There is no OpenGL support in your \
		                                  system."));
		return;
	}
	/*
	Check wether we have GLSL support
	This should be done in here (initializeGL) and not from the constructor,
	as the OpenGL environment is created here.
	*/
	else if(!GLSLUtils::checkSupport(false))
	{
		QMessageBox::critical(0,QObject::tr("This beautiful program"),
		                      QObject::tr("There is no GLSL support in your"
		                                  " system"));
	}

	glEnable(GL_DEPTH_TEST);

	glShadeModel(GL_SMOOTH);
	glEnable(GL_NORMALIZE);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glHint(GL_POLYGON_SMOOTH, GL_NICEST);
	glEnable(GL_POLYGON_SMOOTH);
	glHint(GL_LINE_SMOOTH, GL_NICEST);
	//glEnable(GL_LINE_SMOOTH);

	glLineWidth(0);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	setAutoBufferSwap(false);

	// Enable textures
	glEnable(GL_TEXTURE_2D);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
	lastPos = event->pos();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
	int dx = event->x() - lastPos.x();
	int dy = event->y() - lastPos.y();

	if (event->buttons() & Qt::LeftButton)
	{
		setXRotation(xRot + 8 * dy);
		setYRotation(yRot + 8 * dx);
	}
	else if (event->buttons() & Qt::RightButton)
	{
		setXRotation(xRot + 8 * dy);
		setZRotation(zRot + 8 * dx);
	}
	lastPos = event->pos();
	update();
}

void GLWidget::setShaders()
{
	QString infoLog = tr("No errors found");

	v = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	f = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

	QFile vertexShader(":shaders/shader.vert");
	vertexShader.open(QIODevice::ReadOnly);

	QFile fragmentShader;

	if(useHalfway)
		fragmentShader.setFileName(":/shaders/shaderhw.frag");
	else
		fragmentShader.setFileName(":/shaders/shader.frag");

	fragmentShader.open(QIODevice::ReadOnly);

	QByteArray vertexByteArray = vertexShader.readAll();
	QByteArray fragmentByteArray = fragmentShader.readAll();

	const char * vv = vertexByteArray.data();
	const char * ff = fragmentByteArray.data();

	glShaderSourceARB(v, 1, &vv,NULL);
	glShaderSourceARB(f, 1, &ff,NULL);

	glCompileShaderARB(v);
	if(GLSLUtils::checkCompileError(v, infoLog, true))
	{
		QMessageBox::critical(this,tr("This beautiful program"),
		                      tr("The vertex shader failed to compile"
		                         " with error:\n\n") + infoLog);
		return;
	}

	glCompileShaderARB(f);
	if(GLSLUtils::checkCompileError(f, infoLog, true))
	{
		QMessageBox::critical(this, tr("This beautiful program"),
		                      tr("The <b>fragment</b> shader failed to compile"
                             " with error:\n\n") + infoLog);
		return;
	}

	p = glCreateProgramObjectARB();
	glAttachObjectARB(p,v);
	glAttachObjectARB(p,f);

	glLinkProgramARB(p);
	if(GLSLUtils::checkLinkError(p, infoLog, true))
	{
		QMessageBox::critical(this, tr("This beautiful program"),
		                      tr("The program failed to link"
		                         " with error:\n\n") + infoLog);
		return;
	}

	glUseProgramObjectARB(p);
}

void GLWidget::unsetShaders()
{
	glDetachObjectARB(p, v);
	glDetachObjectARB(p, f);
	glDeleteObjectARB(v);
	glDeleteObjectARB(f);
	glDeleteObjectARB(p);
}

QVector<double> * GLWidget::getNormalVectors(QVector<double> * vertex,
                                             QVector<double> * normals)
{
	QVector<double> * nv = new QVector<double>;
	double x(0.0), y(0.0), z(0.0);
	double norm(0.0);

	for(int i=0; i < vertex->size()/3; i++)
	{
		x = normals->at(3*1);
		y = normals->at((3*i)+1);
		z = normals->at((3*i)+2);

		norm = sqrt(x*x + y*y + z*z);
		norm *=10;

		nv->append(vertex->at(3*i));
		nv->append(vertex->at((3*i)+1));
		nv->append(vertex->at((3*i)+2));
		nv->append(vertex->at(3*i) + x/norm);
		nv->append(vertex->at((3*i)+1) + y/norm);
		nv->append(vertex->at((3*i)+2) + z/norm);
	}

	return nv;
}

void GLWidget::normalizeAngle(int * angle)
{
	while (*angle < 0)
		*angle += 360 * 16;
	while (*angle > 360 * 16)
		*angle -= 360 * 16;
}

Point4F GLWidget::bicubicCalculation(const ConnectivityMatrix & conMat,
                                     const int firstRow, const int firstColumn,
                                     const double u, const double v)
{
	double a(0.0), b(0.0), c(0.0), d(0.0);
	double i(0.0), j(0.0), k(0.0), l(0.0);
	if(currentMethod == Constants::Bezier)
	{
		a = -(u*u*u) + 3*u*u - 3*u +1;
		b = 3*u*u*u - 6*u*u + 3*u;
		c = -3*u*u*u + 3*u*u;
		d = u*u*u;

		i = -(v*v*v) + 3*v*v - 3*v +1;
		j = 3*v*v*v - 6*v*v + 3*v;
		k = -3*v*v*v + 3*v*v;
		l = v*v*v;
	}
	else if(currentMethod == Constants::BSplines)
	{
		a = (-(u*u*u) + 3*u*u - 3*u +1)/6.0;
		b = (3*u*u*u - 6*u*u +4)/6.0;
		c = (-3*u*u*u + 3*u*u + 3*u +1)/6.0;
		d = (u*u*u)/6.0;

		i = (-(v*v*v) + 3*v*v - 3*v +1)/6.0;
		j = (3*v*v*v - 6*v*v +4)/6.0;
		k = (-3*v*v*v + 3*v*v + 3*v +1)/6.0;
		l = (v*v*v)/6.0;
	}

	Point4F e = a * conMat.getPoint(firstRow, firstColumn) +
	            b * conMat.getPoint(firstRow + 1, firstColumn) +
	            c * conMat.getPoint(firstRow + 2, firstColumn) +
	            d * conMat.getPoint(firstRow + 3, firstColumn);
	Point4F f = a * conMat.getPoint(firstRow, firstColumn + 1) +
	            b * conMat.getPoint(firstRow + 1, firstColumn + 1) +
	            c * conMat.getPoint(firstRow + 2, firstColumn + 1) +
	            d * conMat.getPoint(firstRow + 3, firstColumn + 1);
	Point4F g = a * conMat.getPoint(firstRow, firstColumn + 2) +
	            b * conMat.getPoint(firstRow + 1, firstColumn + 2) +
	            c * conMat.getPoint(firstRow + 2, firstColumn + 2) +
	            d * conMat.getPoint(firstRow + 3, firstColumn + 2);
	Point4F h = a * conMat.getPoint(firstRow, firstColumn + 3) +
	            b * conMat.getPoint(firstRow + 1, firstColumn + 3) +
	            c * conMat.getPoint(firstRow + 2, firstColumn + 3) +
	            d * conMat.getPoint(firstRow + 3, firstColumn + 3);

	double x = e.x()*i + f.x()*j + g.x()*k + h.x()*l;
	double y = e.y()*i + f.y()*j + g.y()*k + h.y()*l;
	double z = e.z()*i + f.z()*j + g.z()*k + h.z()*l;

	return Point4F(x, y, z);
}

void GLWidget::calculateSurface()
{
	// Clear all the previous data
	qDeleteAll(holdersList.begin(), holdersList.end());
	holdersList.clear();

	DataHolder * holder = 0;
	ConnectivityMatrix * conMatrix = 0;

	if(currentFigure == Constants::Patch)
		conMatrix = &patchConMat;
	else if(currentFigure == Constants::Tube)
		conMatrix = &tubeConMat;
	else if(currentFigure == Constants::SemiTube)
		conMatrix = &semiTubeConMat;
	else if(currentFigure == Constants::Torus)
		conMatrix = &torusConMat;

	double uStep = 0.0;
	double vStep = 0.0;

	/*
	  We must iterate between 0 and rows-4 and columns-4, as the algorythm takes
	  4 points in each run.
	*/
	for(int row=0; row <= conMatrix->getRows() - 4; row++)
	{
		for(int column=0; column <= conMatrix->getColumns() - 4; column++)
		{
			holder = new DataHolder(numberOfSteps + 1, numberOfSteps + 1);

			for(int u=0; u <= numberOfSteps; u++)
			{
				uStep = (double)u/(double)numberOfSteps;

				for(int v=0; v <= numberOfSteps; v++)
				{
					vStep = (double)v/(double)numberOfSteps;
					holder->addData(u, v,
					                bicubicCalculation(*conMatrix, row, column,
					                                   uStep, vStep));
				}

			}
			holdersList.append(holder);
		}
	}
}




