/***************************************************************************
 *   Copyright (C) 2005 by                                                 *
 *   Lisandro Damián Nicanor Pérez Meyer - perezmeyer en/at gmail.com      *
 *   Gustavo González - gonzalgustavo en/at gmail.com                      *
 *   Pablo Odorico  pablo.odorico en/at gmail.com                          *
 *   Originally designed for QAntenna                                      *
 *   http://qantenna.sourceforge.net/                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef GLWIDGET_H
#define GLWIDGET_H

// GLEW must be included first or it complains
#include <GL/glew.h>
#include <QtOpenGL/QGLWidget>
#include <QList>
#include <QVector>
#include <QPoint>
#include "constants.h"
#include "connectivitymatrix.h"
#include "patch.h"
#include "tube.h"
#include "semitube.h"
#include "torus.h"
#include "point4f.h"
#include "dataholder.h"

class QPixmap;

/**
	This class inherits from QGLWidget, and it is used to diaplay the OGL view
	in the program. It also takes cares of some events, wich it then passes to
	the class camera.
*/
class GLWidget : public QGLWidget
{
	Q_OBJECT

public:
	GLWidget(QWidget * parent = 0);
	~GLWidget();

public slots:
	/// Sets the axis to be drawn or not
	void setDrawAxis(bool draw = true);
	/// Sets the drawing of the normals
	void setDrawNormals(bool draw);
	/// Set the drawing of the wireframe
	void setDrawWireframe(bool draw);
	/// Set the figure to draw
	void setFigure(Constants::Figures figure);
	/// Set the method to draw
	void setMethod(Constants::Methods method);

	/// Sets the X rotation angle
	void setXRotation(int angle);
	/// Sets the Y rotation angle
	void setYRotation(int angle);
	/// Sets the Z rotation angle
	void setZRotation(int angle);

	/// Set use shaders
	void setUseShaders(int use);
	/// Set use half way vector
	void setUseHalfway(int use);
	/// Set the current texture
	void setTexture(int texture);

	/// Set the X coordinate ot the light position
	void setXLightPosition(int position);
	/// Set the Y coordinate ot the light position
	void setYLightPosition(int position);
	/// Set the Z coordinate ot the light position
	void setZLightPosition(int position);

signals:
	/// Emits the light X position
	void lightXPositionChanged(int position);
	/// Emits the light Y position
	void lightYPositionChanged(int position);
	/// Emits the light Z position
	void lightZPositionChanged(int position);

protected:
	/**
	Renders the OpenGL scene. Gets called whenever the widget needs to
	be updated.
	*/
	void paintGL();
	/**
	Sets up the OpenGL viewport, projection, etc. Gets called whenever the
	widget has been resized.
	*/
	void resizeGL(int width, int height);
	/**
	Sets up the OpenGL rendering context, defines display lists, etc.\n
	Gets called once before the first time resizeGL() or paintGL() is called.
	*/
	void initializeGL();
	/// What to do if a button of the mouse is pressed
	void mousePressEvent(QMouseEvent *event);
	/// What to do if the mouse is moved while a button is being pressed
	void mouseMoveEvent(QMouseEvent *event);

private:
	/// Normalize the rotation angle
	void normalizeAngle(int * angle);
	/// Bicubic calculation
	Point4F bicubicCalculation(const ConnectivityMatrix & conMat,
	                           const int firstRow, const int firstColumn,
	                           const double u, const double v);

	/// Calculate the surface
	void calculateSurface();

	/// Sets the shaders to use
	void setShaders();

	/// Unsets the shaders and the program
	void unsetShaders();

	/// Get the normal vectors
	QVector<double> * getNormalVectors(QVector<double> * vertex,
	                                   QVector<double> * normals);

	/// X rotation angle
	int xRot;
	/// Y rotation angle
	int yRot;
	/// Z rotation angle
	int zRot;
	/// Last position of the mouse to calculate rotations
	QPoint lastPos;

	/// State of the axis
	bool drawAxis;

	/// State of the figures normals
	bool drawNormals;

	/// State of the wireframe
	bool drawWireframe;

	/// State of the halfWay vector
	bool useHalfway;

	/// Figure to draw
	int currentFigure;

	/// Method to draw
	int currentMethod;

	/// Step of u and v
	static const int numberOfSteps = 10;

	// Surfaces connectivity matrices
	/// The patch surface connectivity matrix
	Patch patchConMat;
	/// The tube surface connectivity matrix
	Tube tubeConMat;
	/// The semi tube surface connectivity matrix
	SemiTube semiTubeConMat;
	/// Torus surface connectivity matrix
	Torus torusConMat;

	/// Vertex array
	QVector<double> * vertexArray;

	/// Normal array
	QVector<double> * normalArray;

	/// Vectors normal array
	QVector<double> * vectorNormalArray;

	/// List of holders
	QList< DataHolder * > holdersList;

	/// Light position
	float lightPosition[4];

	/// Handles for the vertex shader and fragment shaders and the GLSL program
	GLhandleARB v,f,p;

	/// Texture ID
	GLuint texName;
	/// Current texture
	int currentTexture;
	/// Maximum number of textures
	static const int maxTextures = 16;
};

#endif // GLWIDGET_H

